/*
 * Copyright (c) 2003,2004,2005 William Pitcock
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "client.h"
#include "s_zip.h"
#include "irc_string.h"
#include "packet.h"
#include "s_bsd.h"
#include "s_serv.h"
#include "s_conf.h"
#include "send.h"
#include "memory.h"

#include <string.h>
#include <stdlib.h>

#define ZIP_BUFFER_SIZE         (ZIP_MAXIMUM + READBUF_SIZE)
#define UNZIP_BUFFER_SIZE       6 * ZIP_BUFFER_SIZE
#define ZIP_LEVEL		ConfigFileEntry.compression_level

static char unzipbuf[UNZIP_BUFFER_SIZE];
static char zipbuf[ZIP_BUFFER_SIZE];

int
zip_init(struct Client *client_p)
{
	client_p->zip = (struct Zipdata *)MyMalloc(sizeof(struct Zdata));
	client_p->zip->incount = 0;
	client_p->zip->outcount = 0;

	client_p->zip->in = (z_stream *) MyMalloc(sizeof(z_stream));
	client_p->zip->in->total_in = 0;
	client_p->zip->in->total_out = 0;
	client_p->zip->in->zalloc = (alloc_func) 0;
	client_p->zip->in->zfree = (free_func) 0;
	client_p->zip->in->data_type = Z_ASCII;
	if (inflateInit(client_p->zip->in) != Z_OK) {
		client_p->zip->out = NULL;
		return -1;
	}

	client_p->zip->out = (z_stream *) MyMalloc(sizeof(z_stream));
	client_p->zip->out->total_in = 0;
	client_p->zip->out->total_out = 0;
	client_p->zip->out->zalloc = (alloc_func) 0;
	client_p->zip->out->zfree = (free_func) 0;
	client_p->zip->out->data_type = Z_ASCII;
	if (deflateInit(client_p->zip->out, ZIP_LEVEL) != Z_OK)
		return -1;

	return 0;
}

void
zip_free(struct Client *cptr)
{
	cptr->ziplink = 0;
	if (cptr->zip) {
		if (cptr->zip->in)
			inflateEnd(cptr->zip->in);
		MyFree(cptr->zip->in);
		if (cptr->zip->out)
			deflateEnd(cptr->zip->out);
		MyFree(cptr->zip->out);
		MyFree(cptr->zip);
		cptr->zip = NULL;
	}
}

char *
unzip_packet(struct Client *cptr, char *buffer, int *length)
{
	z_stream *zin = cptr->zip->in;
	int r;
	char *p;

	if (cptr->zip->incount) {
		memcpy((void *)unzipbuf, (void *)cptr->zip->inbuf,
		    cptr->zip->incount);
		zin->avail_out = UNZIP_BUFFER_SIZE - cptr->zip->incount;
		zin->next_out = (Bytef *) (unzipbuf + cptr->zip->incount);
		cptr->zip->incount = 0;
		cptr->zip->inbuf[0] = '\0';	/* again unnecessary but nice for debugger */
	} else {
		if (!buffer) {	/* Sanity test never hurts */
			*length = -1;
			return ((char *)NULL);
		}
		zin->next_in = (Bytef *) buffer;
		zin->avail_in = *length;
		zin->next_out = (Bytef *) unzipbuf;
		zin->avail_out = UNZIP_BUFFER_SIZE;
	}

	switch (r = inflate(zin, Z_NO_FLUSH)) {
	case Z_OK:
		if (zin->avail_in) {
			cptr->zip->incount = 0;

			if (zin->avail_out == 0) {
				sendto_realops_flags(UMODE_SERVNOTICE, L_ALL,
				    "Overflowed unzipbuf increase UNZIP_BUFFER_SIZE");

				if ((zin->next_out[0] == '\n')
				    || (zin->next_out[0] == '\r')) {
					cptr->zip->inbuf[0] = '\n';
					cptr->zip->incount = 1;
				} else {
					for (p = (char *)zin->next_out;
					    p >= unzipbuf;) {
						if ((*p == '\r')
						    || (*p == '\n'))
							break;
						zin->avail_out++;
						p--;
						cptr->zip->incount++;
					}

					sendto_realops_flags(UMODE_SERVNOTICE, L_ALL, "debug: %s",
							unzipbuf);

					if (p == unzipbuf) {
						cptr->zip->incount = 0;
						cptr->zip->inbuf[0] = '\0';	/* only for debugger */
						*length = -1;
						return ((char *)NULL);
					}

					p++;
					cptr->zip->incount--;
					memcpy((void *)cptr->zip->inbuf,
					    (void *)p, cptr->zip->incount);
				}
			} else {
				*length = -1;
				return ((char *)NULL);
			}
		}

		*length = UNZIP_BUFFER_SIZE - zin->avail_out;
		return unzipbuf;

	case Z_BUF_ERROR:
		if (zin->avail_out == 0) {
			sendto_realops_flags(UMODE_SERVNOTICE, L_ALL,
			    "inflate() returned Z_BUF_ERROR: %s",
			    (zin->msg) ? zin->msg : "?");
			*length = -1;
		}
		break;

	case Z_DATA_ERROR:
		if ((cptr->ziplink == 1) && !strncmp("ERROR ", buffer, 6)) {
			return buffer;
		}

		/*
		 * no break 
		 */

	default:
		sendto_realops_flags(UMODE_SERVNOTICE, L_ALL,
		    "inflate() error(%d): %s: %s", r, zError(r),
		    (zin->msg) ? zin->msg : "?");
		*length = -1;
		break;
	}
	return ((char *)NULL);
}

char *
zip_buffer(struct Client *cptr, char *buffer, int *length, int flush)
{
	z_stream *zout = cptr->zip->out;
	int r;

	if (buffer) {
		/*
		 * concatenate buffer in cptr->zip->outbuf 
		 */
		memcpy((void *)(cptr->zip->outbuf + cptr->zip->outcount),
		    (void *)buffer, *length);
		cptr->zip->outcount += *length;
	}
	*length = 0;

	if (!flush && ((cptr->zip->outcount < ZIP_MINIMUM) ||
		((cptr->zip->outcount < (ZIP_MAXIMUM - BUFSIZE)))))
		return ((char *)NULL);

	zout->next_in = (Bytef *) cptr->zip->outbuf;
	zout->avail_in = cptr->zip->outcount;
	zout->next_out = (Bytef *) zipbuf;
	zout->avail_out = ZIP_BUFFER_SIZE;

	switch (r = deflate(zout, Z_PARTIAL_FLUSH)) {
	case Z_OK:
		if (zout->avail_in) {
			/*
			 * can this occur?? I hope not... 
			 */
			sendto_realops_flags(UMODE_SERVNOTICE, L_ALL,
			    "deflate() didn't process all available data!");
		}
		cptr->zip->outcount = 0;
		*length = ZIP_BUFFER_SIZE - zout->avail_out;
		return zipbuf;

	default:		/* error ! */
		sendto_realops_flags(UMODE_SERVNOTICE, L_ALL,
		    "deflate() error(%d): %s: %s", r, zError(r),
		    (zout->msg) ? zout->msg : "?");
		*length = -1;
		break;
	}
	return ((char *)NULL);
}
